#include "utils.h"
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <linux/if.h>
#include <linux/netlink.h>
#include <linux/rtnetlink.h>
#include <linux/route.h>
#include <linux/if_arp.h>
#include <stdio.h>
#include <errno.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <ctype.h>
#include <openssl/evp.h>
#include <openssl/buffer.h>


static int get_iface_param(const char *ifname, void *data, int op)
{
    int fd = socket(AF_INET, SOCK_DGRAM, 0);
    if(fd == -1) {
	sp_log(LOG_MAIN, LOG_ERR, "cannot create socket: %s", strerror(errno));
	return -1;
    }

    struct ifreq ifr;
    ifr.ifr_addr.sa_family = AF_INET;
    strncpy(ifr.ifr_name, ifname, IFNAMSIZ-1);
    if(ioctl(fd, op, &ifr) == -1) {
	close(fd);
	return -1;
    }
    close(fd);

    struct sockaddr_in *addr = (struct sockaddr_in *)&ifr.ifr_addr;
    if(op == SIOCGIFHWADDR)
	memcpy(data, ifr.ifr_hwaddr.sa_data, 6);
    else
	*(uint32_t *)data = addr->sin_addr.s_addr;
    return 0;
}

/**
 * \brief Get interface MAC address
 */
int get_iface_mac(const char *ifname, uint8_t *mac)
{
    return get_iface_param(ifname, mac, SIOCGIFHWADDR);
}

/**
 * \brief Get interface IP address
 */
int get_iface_ip(const char *ifname, uint32_t *ip)
{
    return get_iface_param(ifname, ip, SIOCGIFADDR);
}

int get_iface_mask(const char *ifname, uint32_t *msk)
{
    return get_iface_param(ifname, msk, SIOCGIFNETMASK);
}

int get_iface_peer(const char *ifname, uint32_t *ip)
{
    return get_iface_param(ifname, ip, SIOCGIFDSTADDR);
}

static int set_iface_param(const char *ifname, uint32_t ip, int op)
{
    int fd = socket(AF_INET, SOCK_DGRAM, 0);
    if(fd == -1) {
	sp_log(LOG_MAIN, LOG_ERR, "cannot create socket: %s", strerror(errno));
	return -1;
    }

    struct ifreq ifr;
    ifr.ifr_addr.sa_family = AF_INET;
    struct sockaddr_in *addr = (struct sockaddr_in *)&ifr.ifr_addr;
    addr->sin_addr.s_addr = ip;
    strncpy(ifr.ifr_name, ifname, IFNAMSIZ-1);
    if(ioctl(fd, op, &ifr) == -1) {
	close(fd);
	return -1;
    }
    close(fd);

    return 0;
}

int set_iface_ip(const char *ifname, uint32_t ip)
{
    return set_iface_param(ifname, ip, SIOCSIFADDR);
}

int set_iface_mask(const char *ifname, uint32_t msk)
{
    return set_iface_param(ifname, msk, SIOCSIFNETMASK);
}

struct in6_ifreq {
    struct in6_addr ifr6_addr;
    __u32 ifr6_prefixlen;
    unsigned int ifr6_ifindex;
};

int set_iface_ip6(const char *ifname, void *ip, int prefixlen)
{
    int fd = socket(AF_INET6, SOCK_DGRAM, 0);
    if(fd == -1)
	return -1;

    struct ifreq ifr;
    strncpy(ifr.ifr_name, ifname, IFNAMSIZ-1);

    if(ioctl(fd, SIOGIFINDEX, &ifr) < 0) {
	close(fd);
	return -1;
    }

    struct in6_ifreq ifr6;
    memcpy(&ifr6.ifr6_addr, ip, sizeof(struct in6_addr));
    ifr6.ifr6_ifindex = ifr.ifr_ifindex;
    ifr6.ifr6_prefixlen = prefixlen;

    if(ioctl(fd, SIOCSIFADDR, &ifr6) == -1) {
	close(fd);
	return -1;
    }
    close(fd);
    return 0;
}

int iface_up(const char *ifname)
{
    int fd = socket(AF_INET, SOCK_DGRAM, 0);
    if(fd == -1) {
	sp_log(LOG_MAIN, LOG_ERR, "cannot create socket: %s", strerror(errno));
	return -1;
    }

    struct ifreq ifr;
    memset(&ifr, 0, sizeof(ifr));
    strncpy(ifr.ifr_name, ifname, IFNAMSIZ-1);
    if(ioctl(fd, SIOCGIFFLAGS, &ifr) == -1) {
	sp_log(LOG_MAIN, LOG_ERR, "ioctl: %s", strerror(errno));
	close(fd);
	return -1;
    }
    ifr.ifr_flags |= IFF_UP;
    if(ioctl(fd, SIOCSIFFLAGS, &ifr) == -1) {
	sp_log(LOG_MAIN, LOG_ERR, "ioctl: %s", strerror(errno));
	close(fd);
	return -1;
    }
    close(fd);
    return 0;
}

int get_route(uint32_t ip, uint32_t & res, char * ifname)
{
    static unsigned int sequence;

    struct {
	struct nlmsghdr  nh;
	struct rtmsg     msg;
	char             attrbuf[512];
    } req;

    memset(&req, 0, sizeof(req));
    req.nh.nlmsg_len = NLMSG_LENGTH(sizeof(struct rtmsg));
    req.nh.nlmsg_flags = NLM_F_REQUEST;
    req.nh.nlmsg_type = RTM_GETROUTE;
    req.nh.nlmsg_seq = ++sequence;
    req.nh.nlmsg_pid = 0;
    req.msg.rtm_family = AF_INET;
    req.msg.rtm_dst_len = 32;
    req.msg.rtm_table = RT_TABLE_DEFAULT;
    req.msg.rtm_type = RTN_UNICAST;
    req.msg.rtm_protocol = RTPROT_UNSPEC;
    req.msg.rtm_scope = RT_SCOPE_UNIVERSE;

    struct rtattr *rta = (struct rtattr *)(((char *) &req) + NLMSG_ALIGN(req.nh.nlmsg_len));
    rta->rta_type = RTA_DST;
    rta->rta_len = RTA_LENGTH(sizeof(ip));
    req.nh.nlmsg_len = NLMSG_ALIGN(req.nh.nlmsg_len) + RTA_LENGTH(sizeof(ip));
    memcpy(RTA_DATA(rta), &ip, sizeof(ip));

    int rtnetlink_socket = socket(AF_NETLINK, SOCK_DGRAM, NETLINK_ROUTE);
    if(rtnetlink_socket == -1) {
	sp_log(LOG_MAIN, LOG_ERR, "cannot create socket: %s", strerror(errno));
	return -1;
    }

    int len = send(rtnetlink_socket, &req, req.nh.nlmsg_len, 0);
    if(len == -1)
	sp_log(LOG_MAIN, LOG_ERR, "cannot send: %s", strerror(errno));

    len = recv(rtnetlink_socket, &req, sizeof(req), 0);
    close(rtnetlink_socket);
    if(len == -1) {
	sp_log(LOG_MAIN, LOG_ERR, "recv(): %s", strerror(errno));
	return -1;
    }

    if( !NLMSG_OK(&req.nh, len) || len < (int)NLMSG_LENGTH(sizeof(struct rtmsg)) ) {
	sp_log(LOG_MAIN, LOG_ERR, "invalid netlink message data");
	return -1;
    }
    if(req.nh.nlmsg_seq != sequence) {
	sp_log(LOG_MAIN, LOG_ERR, "unexpected sequence number");
	return -1;
    }
    if(req.nh.nlmsg_type == NLMSG_ERROR) {
	struct nlmsgerr *err = (struct nlmsgerr *)&req.msg;
	sp_log(LOG_MAIN, LOG_ERR, "kernel returned error %d (%s)", err->error, strerror(-err->error));
	return -1;
    }
    if(req.nh.nlmsg_type != RTM_NEWROUTE) {
	sp_log(LOG_MAIN, LOG_ERR, "unexpected message type %d", req.nh.nlmsg_type);
	return -1;
    }

    rta = (struct rtattr *)(((char *) &req) + NLMSG_LENGTH(sizeof(struct rtmsg)));
    int attrlen = req.nh.nlmsg_len - NLMSG_LENGTH(sizeof(struct rtmsg));
    uint32_t gw;
    int ifindex;
    bool have_gw = false;
    bool have_oif = false;

    while(attrlen > 0) {
	if( RTA_OK(rta, attrlen) ) {
	    uint8_t *data = (uint8_t*)RTA_DATA(rta);

	    switch(rta->rta_type) {
	    case RTA_GATEWAY:
		if(rta->rta_len == RTA_LENGTH(sizeof(gw))) {
		    memcpy(&gw, data, sizeof(gw));
		    have_gw = true;
		} else {
		    sp_log(LOG_MAIN, LOG_ERR, "RTA_GATEWAY: invalid attribute size %d", rta->rta_len);
		}
		break;
	    case RTA_OIF:
		if(rta->rta_len == RTA_LENGTH(sizeof(ifindex))) {
		    memcpy(&ifindex, data, sizeof(ifindex));
		    have_oif = true;
		} else {
		    sp_log(LOG_MAIN, LOG_ERR, "RTA_OIF: invalid attribute size %d", rta->rta_len);
		}
		break;
	    }

	    rta = RTA_NEXT(rta, attrlen);
	} else {
	    sp_log(LOG_MAIN, LOG_ERR, "RTA_OK is 0");
	    return -1;
	}
    }

    if(ifname && have_oif) {
	int fd = socket(AF_INET, SOCK_DGRAM, 0);
	if(fd == -1) {
	    sp_log(LOG_MAIN, LOG_ERR, "cannot create socket: %s", strerror(errno));
	    return -1;
	}

	struct ifreq ifr;
	ifr.ifr_ifindex = ifindex;
	if(ioctl(fd, SIOCGIFNAME, &ifr) < 0) {
	    close(fd);
	    sp_log(LOG_MAIN, LOG_ERR, "get_route(): cannot get interface name: %s", strerror(errno));
	    return -1;
	}
	close(fd);
	strncpy(ifname, ifr.ifr_name, IFNAMSIZ-1);
    }

    res = have_gw ? gw : ip;
    return 0;
}

//int get_route(const struct in6_addr * ip, struct in6_addr * res, char * ifname, struct in6_addr * localaddr)
//{
//    static unsigned int sequence;
//
//    struct {
//	struct nlmsghdr  nh;
//	struct rtmsg     msg;
//	char             attrbuf[512];
//    } req;
//
//    memset(&req, 0, sizeof(req));
//    req.nh.nlmsg_len = NLMSG_LENGTH(sizeof(struct rtmsg));
//    req.nh.nlmsg_flags = NLM_F_REQUEST;
//    req.nh.nlmsg_type = RTM_GETROUTE;
//    req.nh.nlmsg_seq = ++sequence;
//    req.nh.nlmsg_pid = 1;
//    req.msg.rtm_family = AF_INET6;
//    req.msg.rtm_dst_len = 128;
//    req.msg.rtm_table = RT_TABLE_DEFAULT;
//    req.msg.rtm_type = RTN_UNICAST;
//    req.msg.rtm_protocol = RTPROT_UNSPEC;
//    req.msg.rtm_scope = RT_SCOPE_UNIVERSE;
//
//    struct rtattr *rta = (struct rtattr *)(((char *) &req) + NLMSG_ALIGN(req.nh.nlmsg_len));
//    rta->rta_type = RTA_DST;
//    rta->rta_len = RTA_LENGTH(sizeof(struct in6_addr));
//    req.nh.nlmsg_len = NLMSG_ALIGN(req.nh.nlmsg_len) + RTA_LENGTH(sizeof(struct in6_addr));
//    memcpy(RTA_DATA(rta), ip, sizeof(struct in6_addr));
//
//    int rtnetlink_socket = socket(AF_NETLINK, SOCK_DGRAM, NETLINK_ROUTE);
//    if(rtnetlink_socket == -1) {
//	sp_log(LOG_MAIN, LOG_ERR, "cannot create socket: %s", strerror(errno));
//	return -1;
//    }
//
//    int len = send(rtnetlink_socket, &req, req.nh.nlmsg_len, 0);
//    if(len == -1)
//	sp_log(LOG_MAIN, LOG_ERR, "cannot send: %s", strerror(errno));
//
//    len = recv(rtnetlink_socket, &req, sizeof(req), 0);
//    close(rtnetlink_socket);
//    if(len == -1) {
//	sp_log(LOG_MAIN, LOG_ERR, "recv(): %s", strerror(errno));
//	return -1;
//    }
//
//    if( !NLMSG_OK(&req.nh, len) || len < (int)NLMSG_LENGTH(sizeof(struct rtmsg)) ) {
//	sp_log(LOG_MAIN, LOG_ERR, "invalid netlink message data");
//	return -1;
//    }
//    if(req.nh.nlmsg_seq != sequence) {
//	sp_log(LOG_MAIN, LOG_ERR, "unexpected sequence number");
//	return -1;
//    }
//    if(req.nh.nlmsg_type == NLMSG_ERROR) {
//	struct nlmsgerr *err = (struct nlmsgerr *)&req.msg;
//	sp_log(LOG_MAIN, LOG_ERR, "kernel returned error %d (%s)", err->error, strerror(-err->error));
//	return -1;
//    }
//    if(req.nh.nlmsg_type != RTM_NEWROUTE) {
//	sp_log(LOG_MAIN, LOG_ERR, "unexpected message type %d", req.nh.nlmsg_type);
//	return -1;
//    }
//
//    rta = (struct rtattr *)(((char *) &req) + NLMSG_LENGTH(sizeof(struct rtmsg)));
//    int attrlen = req.nh.nlmsg_len - NLMSG_LENGTH(sizeof(struct rtmsg));
//    int ifindex;
//    bool have_gw = false;
//    bool have_oif = false;
//
//    while(attrlen > 0) {
//	if( RTA_OK(rta, attrlen) ) {
//	    uint8_t *data = (uint8_t*)RTA_DATA(rta);
//
//	    switch(rta->rta_type) {
//	    case RTA_GATEWAY:
//		if(rta->rta_len == RTA_LENGTH(sizeof(struct in6_addr))) {
//		    memcpy(res, data, sizeof(struct in6_addr));
//		    have_gw = true;
//		} else {
//		    sp_log(LOG_MAIN, LOG_ERR, "RTA_GATEWAY: invalid attribute size %d", rta->rta_len);
//		}
//		break;
//	    case RTA_OIF:
//		if(rta->rta_len == RTA_LENGTH(sizeof(ifindex))) {
//		    memcpy(&ifindex, data, sizeof(ifindex));
//		    have_oif = true;
//		} else {
//		    sp_log(LOG_MAIN, LOG_ERR, "RTA_OIF: invalid attribute size %d", rta->rta_len);
//		}
//		break;
//	    case RTA_PREFSRC:
//		if(localaddr && rta->rta_len == RTA_LENGTH(sizeof(struct in6_addr))) {
//		    memcpy(localaddr, data, sizeof(struct in6_addr));
//		}
//		break;
//	    }
//
//	    rta = RTA_NEXT(rta, attrlen);
//	} else {
//	    sp_log(LOG_MAIN, LOG_ERR, "RTA_OK is 0");
//	    return -1;
//	}
//    }
//
//    if(ifname && have_oif) {
//	int fd = socket(AF_INET, SOCK_DGRAM, 0);
//	if(fd == -1) {
//	    sp_log(LOG_MAIN, LOG_ERR, "cannot create socket: %s", strerror(errno));
//	    return -1;
//	}
//
//	struct ifreq ifr;
//	ifr.ifr_ifindex = ifindex;
//	if(ioctl(fd, SIOCGIFNAME, &ifr) < 0) {
//	    close(fd);
//	    sp_log(LOG_MAIN, LOG_ERR, "get_route(): cannot get interface name: %s", strerror(errno));
//	    return -1;
//	}
//	close(fd);
//	strncpy(ifname, ifr.ifr_name, IFNAMSIZ-1);
//    }
//
//    if(!have_gw)
//	memcpy(res, ip, sizeof(struct in6_addr));
//
//    return 0;
//}

int getMac(const char *dev, uint32_t ip, unsigned char *mac)
{
    // default Mac to zero, in case we return an error
    memset(mac, 0, 6);

    int sockfd;
    if((sockfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
	sp_log(LOG_MAIN, LOG_ERR, "socket(): %s", strerror(errno));
	return -1;
    }

    // get interface address
    struct ifreq ifr;
    ifr.ifr_addr.sa_family = AF_INET;
    strncpy(ifr.ifr_name, dev, IFNAMSIZ-1);
    if(ioctl(sockfd, SIOCGIFADDR, &ifr) == -1) {
	sp_log(LOG_MAIN, LOG_ERR, "socket(): %s", strerror(errno));
	close(sockfd);
	return -1;
    }
    struct sockaddr_in *addrp = (struct sockaddr_in *)&ifr.ifr_addr;
    if(ip == addrp->sin_addr.s_addr) {
	// Запрашивают MAC нашего собственного адреса!
	close(sockfd);
	return get_iface_mac(dev, mac);
    }

    struct sockaddr addr;
    memset(&addr, 0, sizeof(addr));
    addr.sa_family = AF_INET;
    memcpy(&addr.sa_data[2], &ip, 4);

    struct arpreq myreq;
    memset(&myreq, 0, sizeof(myreq));
    memcpy(&myreq.arp_pa, &addr, sizeof(addr));
    memcpy(&myreq.arp_dev, dev, sizeof(myreq.arp_dev));

    // try to get an ARP mapping (if available)
    if(ioctl(sockfd, SIOCGARP, (char *) &myreq) < 0) {
	sp_log(LOG_MAIN, LOG_ERR, "ioctl(): %s", strerror(errno));
	close(sockfd);
	return -1;
    }

    memcpy(mac, myreq.arp_ha.sa_data, 6);

    close(sockfd);
    return 0;
}

#define BUFSIZE 8192

static int readNlSock(int sockFd, char *bufPtr, unsigned int seqNum, unsigned int pId)
{
    struct nlmsghdr *nlHdr;
    int msgLen = 0;

    do
    {
	/* Recieve response from the kernel */
	int readLen = recv(sockFd, bufPtr, BUFSIZE - msgLen, 0);
	if(readLen < 0) {
	    return -1;
	}

	nlHdr = (struct nlmsghdr *)bufPtr;

	/* Check if the header is valid */
	if(!NLMSG_OK(nlHdr, (unsigned int)readLen) || (nlHdr->nlmsg_type == NLMSG_ERROR)) {
	    return -1;
	}

	/* Check if the its the last message */
	if(nlHdr->nlmsg_type == NLMSG_DONE) {
	    break;
	} else {
	    /* Else move the pointer to buffer appropriately */
	    bufPtr += readLen;
	    msgLen += readLen;
	}

	/* Check if its a multi part message */
	if((nlHdr->nlmsg_flags & NLM_F_MULTI) == 0) {
	    /* return if its not */
	    break;
	}
    } while((nlHdr->nlmsg_seq != seqNum) || (nlHdr->nlmsg_pid != pId));

    return msgLen;
}

/* parse the route info returned */
static bool parseRoutes(const struct nlmsghdr *nlHdr, uint32_t *gw)
{
    const struct rtmsg *rtMsg;
    const struct rtattr *rtAttr;
    int rtLen;

    rtMsg = (struct rtmsg *)NLMSG_DATA(nlHdr);

    /* If the route is not for AF_INET or does not belong to main routing table then return. */
    if(rtMsg->rtm_family != AF_INET || rtMsg->rtm_table != RT_TABLE_MAIN)
	return false;

    /* get the rtattr field */
    rtAttr = (struct rtattr *)RTM_RTA(rtMsg);
    rtLen = RTM_PAYLOAD(nlHdr);

    for(;RTA_OK(rtAttr,rtLen);rtAttr = RTA_NEXT(rtAttr,rtLen))
    {
	if(rtAttr->rta_type == RTA_GATEWAY) {
	    memcpy(gw, RTA_DATA(rtAttr), sizeof(*gw));
	    return true;
	}
    }

    return false;
}

int get_default_gw(uint32_t *res)
{
    static unsigned int sequence;

    struct {
	struct nlmsghdr  nh;
	struct rtmsg     msg;
	char             attrbuf[512];
    } req;

    memset(&req, 0, sizeof(req));
    req.nh.nlmsg_len = NLMSG_LENGTH(sizeof(struct rtmsg));
    req.nh.nlmsg_flags = NLM_F_REQUEST | NLM_F_DUMP;
    req.nh.nlmsg_type = RTM_GETROUTE;
    req.nh.nlmsg_seq = ++sequence;
    req.nh.nlmsg_pid = getpid();

    int rtnetlink_socket = socket(AF_NETLINK, SOCK_DGRAM, NETLINK_ROUTE);
    if(rtnetlink_socket == -1) {
	return -1;
    }

    /* Send the request */
    int len = send(rtnetlink_socket, &req, req.nh.nlmsg_len, 0);
    if(len == -1) {
	close(rtnetlink_socket);
	return -1;
    }

    char msgBuf[BUFSIZE];
    struct nlmsghdr *nlMsg = (struct nlmsghdr *)msgBuf;

    /* Read the response */
    if((len = readNlSock(rtnetlink_socket, msgBuf, sequence, getpid())) < 0) {
	close(rtnetlink_socket);
	return -1;
    }
    close(rtnetlink_socket);

    /* Parse and print the response */
    for(;NLMSG_OK(nlMsg, (unsigned int)len);nlMsg = NLMSG_NEXT(nlMsg,len)) {
	// Check if default gateway
	if(parseRoutes(nlMsg, res))
	    return 0;
    }

    return -1;
}

int set_default_gw(uint32_t ip)
{
    int fd = socket(AF_INET, SOCK_DGRAM, 0);
    if(fd == -1)
	return -1;

    struct rtentry route;
    memset(&route, 0, sizeof (struct rtentry));
    struct sockaddr_in *dst, *gw, *mask;
    dst = (struct sockaddr_in *) (&(route.rt_dst));
    gw = (struct sockaddr_in *) (&(route.rt_gateway));
    mask = (struct sockaddr_in *) (&(route.rt_genmask));
    dst->sin_family = AF_INET;
    gw->sin_family = AF_INET;
    mask->sin_family = AF_INET;

    // Set up the data for removing the default route
    dst->sin_addr.s_addr = 0;
    gw->sin_addr.s_addr = 0;
    mask->sin_addr.s_addr = 0;
    route.rt_flags = RTF_UP | RTF_GATEWAY;

    // Remove the default route
    ioctl(fd, SIOCDELRT, &route);

    // Set up the data for adding the default route
    dst->sin_addr.s_addr = 0;
    gw->sin_addr.s_addr = ip;
    mask->sin_addr.s_addr = 0;
    route.rt_metric = 1;
    route.rt_flags = RTF_UP | RTF_GATEWAY;

    // Remove this route if it already exists
    ioctl(fd, SIOCDELRT, &route);

    // Add the default route
    if(ioctl(fd, SIOCADDRT, &route) == -1) {
	close(fd);
	return -1;
    }

    close(fd);
    return 0;
}

bool have_global_ip6(const char *ifname)
{
    FILE *f = fopen("/proc/net/if_inet6", "r");
    if(f == NULL) {
	return false;
    }

    bool res = false;
    char addr[36], devname[32];
    int if_idx, plen, scope, flags;
    while (fscanf(f, "%32s %02x %02x %02x %02x %20s\n", addr, &if_idx, &plen, &scope, &flags, devname) != EOF) {
	if(flags & IFA_F_DEPRECATED)
	    continue;

	if((ifname == NULL || strcmp(devname, ifname) == 0) && scope == 0) {
	    res = true;
	    break;
	}
    }

    fclose(f);
    return res;
}

static int hex2dec(char c)
{
    return c <= '9' ? c - '0' : tolower(c) - 'a' + 10;
}

bool get_link_local_ip6(struct in6_addr * ip, const char *ifname, int * prefix_len)
{
    bool found = false;
    char addr[36];
    int plen;

    FILE *f = fopen("/proc/net/if_inet6", "r");
    if(f) {
	char devname[32];
	int if_idx, scope, dad_status;
	while (fscanf(f, "%32s %02x %02x %02x %02x %20s\n", addr, &if_idx, &plen, &scope, &dad_status, devname) != EOF) {
	    if(strcmp(devname, ifname) == 0 && scope == 0x20) {
		found = true;
		break;
	    }
	}

	fclose(f);
    }

    if(found) {
	uint8_t *ipv6 = (uint8_t *)ip;
	for(unsigned int i = 0, p = 0; i < sizeof(struct in6_addr); i++, p += 2) {
	    ipv6[i] = (hex2dec(addr[p]) << 4) | hex2dec(addr[p + 1]);
	}
	if(prefix_len)
	    *prefix_len = plen;
    }
    return found;
}

bool guess_ip6(struct in6_addr * ip, const char *ifname, int * prefix_len)
{
    bool found = false;
    char addr[36];
    int plen;

    FILE *f = fopen("/proc/net/if_inet6", "r");
    if(f) {
	char devname[32];
	int if_idx, scope, flags;
	while (fscanf(f, "%32s %02x %02x %02x %02x %20s\n", addr, &if_idx, &plen, &scope, &flags, devname) != EOF) {
	    if(flags & IFA_F_DEPRECATED)
		continue;

	    if((ifname == NULL || strcmp(devname, ifname) == 0) && scope == 0) {
		found = true;
		break;
	    }
	}

	if(!found) {
	    rewind(f);
	    while (fscanf(f, "%32s %02x %02x %02x %02x %20s\n", addr, &if_idx, &plen, &scope, &flags, devname) != EOF) {
		if(flags & IFA_F_DEPRECATED)
		    continue;

		if((ifname == NULL || strcmp(devname, ifname) == 0) && scope == 0x20) {
		    found = true;
		    break;
		}
	    }
	}

	fclose(f);
    }

    if(found) {
	uint8_t *ipv6 = (uint8_t *)ip;
	for(unsigned int i = 0, p = 0; i < sizeof(struct in6_addr); i++, p += 2) {
	    ipv6[i] = (hex2dec(addr[p]) << 4) | hex2dec(addr[p + 1]);
	}
	if(prefix_len)
	    *prefix_len = plen;
    }
    return found;
}

int drop_iface_ip6(const char *ifname, bool including_link_local)
{
    int fd = socket(AF_INET6, SOCK_DGRAM, 0);
    if(fd == -1)
	return -1;

    FILE *f = fopen("/proc/net/if_inet6", "r");
    if(f == NULL) {
	close(fd);
	return -1;
    }

    char addr[36], devname[32];
    int if_idx, plen, scope, dad_status;
    while (fscanf(f, "%32s %02x %02x %02x %02x %20s\n", addr, &if_idx, &plen, &scope, &dad_status, devname) != EOF) {
	if(strcmp(devname, ifname) != 0 || (scope == 32 && !including_link_local))
	    continue;

	struct in6_ifreq ifr6;
	ifr6.ifr6_ifindex = if_idx;
	ifr6.ifr6_prefixlen = plen;

	// декодируем адрес
	uint8_t *ipv6 = (uint8_t *)&ifr6.ifr6_addr;
	for(unsigned int i = 0, p = 0; i < sizeof(struct in6_addr); i++, p += 2) {
	    ipv6[i] = (hex2dec(addr[p]) << 4) | hex2dec(addr[p + 1]);
	}

	char test[64];
	inet_ntop(AF_INET6, ipv6, test, sizeof(test));
	ioctl(fd, SIOCDIFADDR, &ifr6);
    }

    fclose(f);
    close(fd);
    return 0;
}

int get_local_ipv6(const struct in6_addr * ip, struct in6_addr * res, char * ifname)
{
    static unsigned int sequence;

    struct {
	struct nlmsghdr  nh;
	struct rtmsg     msg;
	char             attrbuf[512];
    } req;

    memset(&req, 0, sizeof(req));
    req.nh.nlmsg_len = NLMSG_LENGTH(sizeof(struct rtmsg));
    req.nh.nlmsg_flags = NLM_F_REQUEST;
    req.nh.nlmsg_type = RTM_GETROUTE;
    req.nh.nlmsg_seq = ++sequence;
    req.nh.nlmsg_pid = 0;
    req.msg.rtm_family = AF_INET6;
    req.msg.rtm_dst_len = 32;
    req.msg.rtm_table = RT_TABLE_DEFAULT;
    req.msg.rtm_type = RTN_UNICAST;
    req.msg.rtm_protocol = RTPROT_UNSPEC;
    req.msg.rtm_scope = RT_SCOPE_UNIVERSE;

    struct rtattr *rta = (struct rtattr *)(((char *) &req) + NLMSG_ALIGN(req.nh.nlmsg_len));
    rta->rta_type = RTA_DST;
    rta->rta_len = RTA_LENGTH(sizeof(struct in6_addr));
    req.nh.nlmsg_len = NLMSG_ALIGN(req.nh.nlmsg_len) + RTA_LENGTH(sizeof(struct in6_addr));
    memcpy(RTA_DATA(rta), ip, sizeof(struct in6_addr));

    int rtnetlink_socket = socket(AF_NETLINK, SOCK_DGRAM, NETLINK_ROUTE);
    if(rtnetlink_socket == -1) {
	sp_log(LOG_MAIN, LOG_ERR, "cannot create socket: %s", strerror(errno));
	return -1;
    }

    int len = send(rtnetlink_socket, &req, req.nh.nlmsg_len, 0);
    if(len == -1)
	sp_log(LOG_MAIN, LOG_ERR, "cannot send: %s", strerror(errno));

    len = recv(rtnetlink_socket, &req, sizeof(req), 0);
    close(rtnetlink_socket);
    if(len == -1) {
	sp_log(LOG_MAIN, LOG_ERR, "recv(): %s", strerror(errno));
	return -1;
    }

    if( !NLMSG_OK(&req.nh, len) || len < (int)NLMSG_LENGTH(sizeof(struct rtmsg)) ) {
	sp_log(LOG_MAIN, LOG_ERR, "invalid netlink message data");
	return -1;
    }
    if(req.nh.nlmsg_seq != sequence) {
	sp_log(LOG_MAIN, LOG_ERR, "unexpected sequence number");
	return -1;
    }
    if(req.nh.nlmsg_type == NLMSG_ERROR) {
	struct nlmsgerr *err = (struct nlmsgerr *)&req.msg;
	sp_log(LOG_MAIN, LOG_ERR, "kernel returned error %d (%s)", err->error, strerror(-err->error));
	return -1;
    }
    if(req.nh.nlmsg_type != RTM_NEWROUTE) {
	sp_log(LOG_MAIN, LOG_ERR, "unexpected message type %d", req.nh.nlmsg_type);
	return -1;
    }

    rta = (struct rtattr *)(((char *) &req) + NLMSG_LENGTH(sizeof(struct rtmsg)));
    int attrlen = req.nh.nlmsg_len - NLMSG_LENGTH(sizeof(struct rtmsg));
    int ifindex;
    bool have_res = false;
    bool have_oif = false;

    while(attrlen > 0) {
	if( RTA_OK(rta, attrlen) ) {
	    uint8_t *data = (uint8_t*)RTA_DATA(rta);

	    switch(rta->rta_type) {
	    case RTA_PREFSRC:
		if(rta->rta_len == RTA_LENGTH(sizeof(struct in6_addr))) {
		    memcpy(res, data, sizeof(struct in6_addr));
		    have_res = true;
		} else {
		    sp_log(LOG_MAIN, LOG_ERR, "RTA_PREFSRC: invalid attribute size %d", rta->rta_len);
		}
		break;
	    case RTA_OIF:
		if(rta->rta_len == RTA_LENGTH(sizeof(ifindex))) {
		    memcpy(&ifindex, data, sizeof(ifindex));
		    have_oif = true;
		} else {
		    sp_log(LOG_MAIN, LOG_ERR, "RTA_OIF: invalid attribute size %d", rta->rta_len);
		}
		break;
	    }

	    rta = RTA_NEXT(rta, attrlen);
	} else {
	    sp_log(LOG_MAIN, LOG_ERR, "RTA_OK is 0");
	    return -1;
	}
    }

    if(ifname && have_oif) {
	int fd = socket(AF_INET, SOCK_DGRAM, 0);
	if(fd == -1) {
	    sp_log(LOG_MAIN, LOG_ERR, "cannot create socket: %s", strerror(errno));
	    return -1;
	}

	struct ifreq ifr;
	ifr.ifr_ifindex = ifindex;
	if(ioctl(fd, SIOCGIFNAME, &ifr) < 0) {
	    close(fd);
	    sp_log(LOG_MAIN, LOG_ERR, "get_local_ipv6(): cannot get interface name: %s", strerror(errno));
	    return -1;
	}
	close(fd);
	strncpy(ifname, ifr.ifr_name, IFNAMSIZ-1);
    }

    if(!have_res) {
	memset(res, 0, sizeof(struct in6_addr));
	res->s6_addr[15] = 1;		// ::1
    }

    return 0;
}

int get_neighbor_mac(const struct in6_addr * ip, uint8_t * mac)
{
    static unsigned int sequence;

    struct Msg {
	struct nlmsghdr	nh;
	struct ndmsg	msg;
	char		data[4];
    };

    struct Msg req;

    memset(&req, 0, sizeof(req));
    req.nh.nlmsg_len = NLMSG_LENGTH(sizeof(struct ndmsg));
    req.nh.nlmsg_flags = NLM_F_REQUEST | NLM_F_ROOT;
    req.nh.nlmsg_type = RTM_GETNEIGH;
    req.nh.nlmsg_seq = ++sequence;
    req.nh.nlmsg_pid = 0;
    req.msg.ndm_family = AF_INET6;
    req.msg.ndm_state = NUD_REACHABLE;

    int rtnetlink_socket = socket(AF_NETLINK, SOCK_DGRAM, NETLINK_ROUTE);
    if(rtnetlink_socket == -1) {
	sp_log(LOG_MAIN, LOG_ERR, "cannot create socket: %s", strerror(errno));
	return -1;
    }

    int len = send(rtnetlink_socket, &req, req.nh.nlmsg_len, 0);
    if(len == -1)
	sp_log(LOG_MAIN, LOG_ERR, "cannot send: %s", strerror(errno));

    char buffer[32768];

    len = recv(rtnetlink_socket, buffer, sizeof(buffer), 0);
    close(rtnetlink_socket);
    if(len == -1) {
	sp_log(LOG_MAIN, LOG_ERR, "recv(): %s", strerror(errno));
	return -1;
    }

    struct Msg * resp = (struct Msg *)buffer;

    const char zeromac[] = { 0, 0, 0, 0, 0, 0 };

    while(len > (int)sizeof(struct Msg)) {
	if( !NLMSG_OK(&resp->nh, len) || len < (int)NLMSG_LENGTH(sizeof(struct ndmsg)) ) {
	    sp_log(LOG_MAIN, LOG_ERR, "invalid netlink message data");
	    return -1;
	}
	if(resp->nh.nlmsg_seq != sequence) {
	    sp_log(LOG_MAIN, LOG_ERR, "unexpected sequence number");
	    return -1;
	}
	if(resp->nh.nlmsg_type == NLMSG_ERROR) {
	    struct nlmsgerr *err = (struct nlmsgerr *)&resp->msg;
	    sp_log(LOG_MAIN, LOG_ERR, "kernel returned error %d (%s)", err->error, strerror(-err->error));
	    return -1;
	}
	if(resp->nh.nlmsg_type != RTM_NEWNEIGH) {
	    sp_log(LOG_MAIN, LOG_ERR, "unexpected message type %d", resp->nh.nlmsg_type);
	    return -1;
	}

	int nlmsg_len = resp->nh.nlmsg_len;
	struct rtattr *rta = (struct rtattr *)(((char *)resp) + NLMSG_LENGTH(sizeof(struct ndmsg)));
	int attrlen = resp->nh.nlmsg_len - NLMSG_LENGTH(sizeof(struct ndmsg));
	bool have_dst = false;
	bool have_mac = false;
	char mactmp[6];

	while(attrlen > 0) {
	    if( RTA_OK(rta, attrlen) ) {
		uint8_t *data = (uint8_t*)RTA_DATA(rta);

		switch(rta->rta_type) {
		case RTA_DST:
		    if(rta->rta_len == RTA_LENGTH(sizeof(struct in6_addr))) {
			have_dst = memcmp(ip, data, sizeof(struct in6_addr)) == 0;
		    } else {
			sp_log(LOG_MAIN, LOG_WARNING, "RTA_DST: invalid attribute size %d", rta->rta_len);
		    }
		    break;

		case RTA_SRC:
		    if(rta->rta_len == RTA_LENGTH(sizeof(mactmp))) {
			memcpy(mactmp, data, sizeof(mactmp));
			have_mac = true;
		    } else {
			sp_log(LOG_MAIN, LOG_WARNING, "RTA_SRC: invalid attribute size %d", rta->rta_len);
		    }
		    break;
		}

		rta = RTA_NEXT(rta, attrlen);
	    } else {
		sp_log(LOG_MAIN, LOG_ERR, "invalid attributes received from kernel socket");
		return -1;
	    }
	}

	if(have_dst && have_mac && memcmp(mactmp, zeromac, sizeof(mactmp)) != 0) {
	    memcpy(mac, mactmp, sizeof(mactmp));
	    return 0;
	}

	len -= NLMSG_ALIGN(nlmsg_len);
	resp = (struct Msg *)(((char*)resp) + NLMSG_ALIGN(nlmsg_len));
    }

    return -1;
}

static size_t calcDecodeLength(const char* b64input)
{
    //Calculates the length of a decoded string
    size_t len = strlen(b64input), padding = 0;

    if (b64input[len-1] == '=' && b64input[len-2] == '=') //last two chars are =
	padding = 2;
    else if (b64input[len-1] == '=') //last char is =
	padding = 1;

    return (len * 3) / 4 - padding;
}
